// ################################################################################
//
//  Light_Tracker v0.0 (R2) candidate, not released
//
//  Released:  14/07/2024
//
//  Author: TechKnowTone
//
// ################################################################################
/*
    TERMS OF USE: This software is furnished "as is", without technical support, and
    with no warranty, expressed or implied, as to its usefulness for any purpose. In
    no event shall the author or copyright holder be liable for any claim, damages,
    or other liability, whether in an action of contract, tort or otherwise, arising
    from, out of or in connection with the software or the use or other dealings in
    the software.

    microcontroller:  ESP32 Dev Module

    This code controls a light tracking robot, for use with a solar cell.

    The code is multi-tasking and does the following:
    * Periodically reads button switches SW0 and SW1
    * Display information like PWM values, ranges and WiFi active flag
    * Able to connect to multiple Wii WEMOS D1 Transcievers

    The Light Tracker
        
    The servo motors must be calibrated during construction, and the values noted
    need to be inserted in the angle definitions included in this code. Functions
    that are commented-out, are not needed but you may find useful to use in your
    code. Enjoy!

    Servo calibration limits updated in this version

    IMPORTANT - Espressif change the way ESP-NOW works from v3.0, which brok the original
    code and caused compiler errors. THis version has been modified to work with v3.0+.
*/
// Declare Libraries
#include <Arduino.h>            // needed when using an ESP32 micro
#include <HardwareSerial.h>     // serial library
#include <Wire.h>               // I2C library
#include "SSD1306Wire.h"        // OLED display library

#include <ESP32Servo.h>         // servo library

#include <esp_now.h>            // ESP-NOW WiFi link library
#include <WiFi.h>               // ESP WiFi library

String Released = "R2";
String Date = "14/07/2024";

// Display Settings
const int I2C_DISPLAY_ADDRESS = 0x3c;
const int SDA_PIN = 21;         // I2C SDA GPIO pin
const int SCL_PIN = 22;         // I2C SCL GPIO pin

// Initialize the oled display for address 0x3c
SSD1306Wire display(I2C_DISPLAY_ADDRESS, SDA_PIN, SCL_PIN);

// Define a task for core 0
TaskHandle_t Core0;

// Define servos constants gained from servo calibration tasks
// All servos are different and leaver arm positions will not match
// These values must be those taken from PWM tests performed on your robot
// If you ever change a servo you will need to re-calibrate those values

// Define general constants
#define BatCal 383.9        // calibrated voltage multiplying factor
#define BatCritical 2534    // critical battery threshold @6.6v, default = 2672, set == 0 to ignore
#define BatMax 3148         // A0 for fully charged battery voltage, == 8.2v
#define BatPin 32           // analog input pin ADC0 on GPIO32
#define BatWarn 2687        // battery threshold of 7.0v gives low warning
#define Flip 110            // excess error causing flip-over
#define PanCtr 1232         // centre Pan servo value in microseconds
#define PanMaxDef 2084      // default maximum servo value, anti-clockwise
#define PanMinDef 500       // defaultminimum servo value, clockwise
#define PanPin 18           // pin connected to Pan servo
#define PanValDef 1500      // default Pan servo angle in microseconds
//#define SensePin0 34        // ADC1_6 analogue pin GPIO34
//#define SensePin1 35        // ADC1_7 analogue pin GPIO35
//#define SensePin2 36        // ADC1_0 analogue pin GPIO36
//#define SensePin3 39        // ADC1_3 analogue pin GPIO39

#define SensePin0 39        // ADC1_6 analogue pin GPIO34
#define SensePin1 34        // ADC1_7 analogue pin GPIO35
#define SensePin2 36        // ADC1_0 analogue pin GPIO36
#define SensePin3 35        // ADC1_3 analogue pin GPIO39

#define sw0Pin 15           // left switch SW0 assigned to GPIO13
#define sw1Pin  2           // right switch SW1 assigned to GPIO15
#define TiltCtr 1346        // centre Tilt servo value in microseconds, vertical
#define TiltMaxDef 2270     // default maximum servo value, tilts towards the display
#define TiltMinDef 530      // defaultminimum servo value, tilts away from the display
#define TiltPin 19          // pin connected to Tilt servo
#define TiltValDef 1500     // default Pan servo angle in microseconds

// create servo instances
Servo servoPan;   // define Light Tracker Pan servo instance
Servo servoTilt;  // define Light Tracker Tilt servos instance

// you need to build a Wii Transceiver and get its MAC address
uint8_t broadcastAddress[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

// 250 byte max for Tx/Rx data
// It will be used to create the Tx/Rx data structures
typedef struct ESPNOW_Buff250 {
    char ESPdata[250];
} ESPNOW_Buff250;

// Create a 250 byte Tx buffer
ESPNOW_Buff250 Tx_Buff;

// Create a 250 byte receive
ESPNOW_Buff250 Rx_Buff;

// Declare and initialise global variables
int16_t Any;                  // any temp value
String Any$;                  // temporary string
int anyInt;                   // any temporary integer value
long AnyLong;                 // any temp value
bool atPanTgt;                // = true if Pan servo is at target value
bool atTiltTgt;               // = true if Tilt servo is at target value
int32_t AvgDiv;               // sensor averaging denominator value
int32_t BatAvg;               // preload averaging function with high value
int32_t BatDist;              // used battery discharge testing
int32_t BatSum;               // preload averaging sum with 20x max value
int32_t BatVol;               // instantaneous battery voltage
bool Border;                  // if true then display has a border
uint8_t Bright;               // FastLED brightness
uint8_t Brightness;           // Display brightness
bool Cal;                     // == true if in calibration mode
byte C_Dn;                    // counter for 'C' button down period
char cmdMode;                 // command mode
int16_t cmdSgn;               // cmdVal sign, 0 or -1
char cmdType;                 // command mode type
int cmdVal;                   // value associated with a cmdType
int16_t CZ;                   // received button values
bool DispBack;                // flag used to draw a back arrow on the display
bool DispChng;                // flag invokes initialisation code
bool DispClr;                 // set == true to clear the memory contents
int16_t DispCnt;              // counter used in display updating
byte DispCX,DispCY;           // display cursor coordinates
int16_t DispDel;              // >0 display delay counter, stops updates over I2C
int16_t DispDelGbl;           // global DispDel offset, used to adjust display frame rates
bool DispDisp;                // set = true to immediately display memory contents
bool DispLock;                // if == true then don't allow display mode changes
byte DispMenu;                // group type, 0 = sensors, 1 = tasks, etc
int16_t DispMode;             // the type of display for a given mode
bool DispMon;                 // == true if connected to OLED Mirror app
int16_t DispNext;             // special temp display mode, default = 0
bool DispNOP;                 // set == true to stop display I2C updates
int16_t DispNow;              // if > 0 display special modes
int16_t DispOveride;          // if DispLock == true then use this overide mode
int16_t DispPnt;              // points at the current display mode
int16_t DispTx;               // counter used with display mirror app
int16_t DmB,DmS,DmX,DmY;      // button, SHIFT and X,Y values associated with a display monitor mouse click
int16_t DM_Max,DM_Min;        // min/max Display Mode pointer values
int16_t ErrDiv;               // integrator error scale factor
bool Echo;                    // if true send print messages
int16_t ErrPan;               // pan sensor difference error
int16_t ErrTilt;              // tilt sensor difference error
bool ESC;                     // true to escape all functions
int16_t ESP_NOW_BA;           // broadcast address pointer, -1 to start with
bool ESP_NOW_Init;            // == true once ESP-NOW has been initiaiised
int16_t JoyX;                 // received value
int16_t JoyY;                 // received value
int JX;                       // global variable used with Joystick app
int JY;                       // global variable used with Joystick app
char keyChar;                 // any keyboard character
int16_t keyVal;               // any keyboard value
int16_t MainTask;             // main task pointer
String myMAC;                 // MAC address if this device
unsigned long next10ms;       // 10ms loop timer
unsigned long next20ms;       // 20ms loop timer
unsigned long next40ms;       // 40ms loop timer
bool PanAtt;                  // = true when Pan servo has been attached
unsigned long PanInterval;    // main loop Pan movement interval in microseconds
int16_t PanMax;               // maximum servo value
unsigned long PanMicros;      // main loop Pan movement timer in microseconds
int16_t PanMin;               // minimum Pan servo value
int16_t PanPause;            // 20ms counter to temporarily pause Pan movement
int16_t PanTgt;               // target Pan servo angle
int16_t PanTgtCnt;           // counter used in automatic Pan servo move to target actions
int PanVal;                   // current Pan servo angle in microseconds
int16_t Ping;                 // Rx 'ping' counter
unsigned long print40ms;      // 40ms variable print timer
int16_t PrintTgt;             // 0 == serial port, 1 == WiFi ESP-NOW
String PrintTx = "";          // printed strings
int16_t PWM_Freq;             // servo PWM frequency, default = 100;
int16_t Rx_Chk;               // Rx checksum
int16_t RxCZ;                 // buffered received CZ value
int16_t RxJoyX;               // buffered received JoyX value
int16_t RxJoyY;               // buffered received JoyY value
int16_t Rx_len;               // length of WiFi received data block
int16_t Rx_Pnt;               // data pointer user by Rx state machine
bool RxRec;                   // set true when a valid frame of data is received
bool RxRecvEvent;             // set == true when a OnDataRecv() call back event has occured
int16_t RxState;              // receiver state machine state
int16_t Rx_Task;              // task pointer user by Rx state machine
int16_t RTxTimeout;           // counter used to auto-reset if WiFi fails
int16_t RxVal;                // value received from serial Rx
byte RxWiFi[6];               // array holding 6 bytes of Wii received data
int16_t Sense0, Sense1,Sense2,Sense3;             // light sensor adjusted readings
int16_t Sense0Avg, Sense1Avg,Sense2Avg,Sense3Avg; // light sensor averaged readings
int16_t Sense0Old, Sense1Old,Sense2Old,Sense3Old; // light sensor averaged readings
int16_t Sense0Raw, Sense1Raw,Sense2Raw,Sense3Raw; // light sensor raw readings
int16_t Sense0Ref, Sense1Ref,Sense2Ref,Sense3Ref; // light sensor reference readings
int Sense0Sum, Sense1Sum,Sense2Sum,Sense3Sum;     // light sensor accumulators readings
bool SenseCal;                // if == true then sensor averaging has occured
int16_t SensePnt;             // pointer to sensor to be read in Core0 task
bool SerialRx;                // == true if any character has been received
int16_t servoDelta;           // temporary delta value used in target movements
bool servoPanEn;              // set = true if Pan servo is attached
int16_t ServoPnt;             // servo target pointer for manual control
bool servoTiltEn;             // set = true if Tilt servo is attached
int16_t SubCnt;               // sub task counter
int16_t SubTask;              // sub task pointer
int16_t sw0Cnt;               // button switch counter
int16_t sw0DwnTime;           // button pressed down time
bool sw0LastState;            // previous state of button switch, HIGH/LOW
bool sw0_Nop;                 // == true to block SW0 read tasks, default = false
bool sw0State;                // state of read button switch pin
int16_t sw0Timer;             // timer used to detemine button sequences
bool sw0WaitHi;               // set == true to wait for a HIGH state
int16_t sw1Cnt;               // button switch counter
int16_t sw1DwnTime;           // button pressed down time
bool sw1LastState;            // previous state of button switch, HIGH/LOW
bool sw1_Nop;                 // == true to block SW1 read tasks, default = false
bool sw1WaitHi;               // set == true to wait for a HIGH state
bool sw1State;                // state of read button switch pin
int16_t sw1Timer;             // timer used to detemine button sequences
int16_t Task10ms;             // 10ms subtask timer
int16_t Task20ms;             // 20ms subtask timer
int16_t Task40ms;             // 40ms subtask timer
bool TEST;                    // normally false, set = true for assembly test mode
bool TiltAtt;                 // = true when Tilt servo has been attached
unsigned long TiltInterval;   // main loop Tilt movement interval in microseconds
int16_t TiltMax;              // maximum Tilt servo value
unsigned long TiltMicros;     // main loop Tilt movement timer in microseconds
int16_t TiltMin;              // minimum Tilt servo value
int16_t TiltPause;            // 20ms counter to temporarily pause Tilt movement
int16_t TiltTgt;              // target Tilt servo angle
int16_t TiltTgtCnt;           // counter used in automatic Tilt servo move to target actions
int TiltVal;                  // current Tilt servo angle in microseconds
int16_t Tx_PingCnt;           // 1 second Tx ping timer
bool USB;                     // == true if initial battery is below 5.5v
int16_t WiFiCntC;             // C button counter for WiFi enable
int16_t WiFiConCnt;           // counter used in connection retry times
bool WiFiConnected;           // == true if a WiFi connection has been established
bool WiFiEn;                  // true when WiFi is enabled with prolonged 'Z' button activity
int16_t WiFiPing;             // Rx counter used to maintain connected state
int16_t WiFiRestCnt;          // counter used to goi to rest if no WiFi demands
bool WiFiRx;                  // true when a block of data has been received
long WiFiRxBytes;             // number of bytes received over the WiFi link
int16_t WiFiRxErr;            // number of bytes received over the WiFi link
bool WiFiRxRec;               // true when a block of data has been successfully received
byte WiFiRxMacAddr[6];        // senders MAC address, which could be broadcast FF:FF:FF:FF:FF:FF
int16_t WiFiTryCnt;           // WiFi try to connect count on error count
int16_t WiFiTryNum;           // WiFi try to connect total count
bool WiFiTryOnce;             // == true if first pass of trying to connect
bool WiFiTx;                  // true when a block of data has been sent successfully
int16_t WiFiTx_CB;            // == 1 call back is clear to send, == -1 resend
int16_t WiFiTx_Chk;           // Tx checksum
int16_t WiFiTx_len;           // >0 when WiFi Tx buffer contains data to be sent
long WiFiTxBytes;             // number of bytes received over the WiFi link
int16_t WiFiTxErr;            // WiFi data error
int16_t WiFiTxErrCnt;         // WiFi Tx error count, leads to disconnection
byte WiFiTxMacAddr[6];        // transmit MAC address, which should match myMAC address
int16_t WiiType;              // device type, 0 == Nunchuk, 1 == Classic
byte Z_Dn;                    // button pressed state; 0 == UP, >= 1 == Down

// ---------------------------------------------------------------------------------

void setup() {
  // Setup code, runs once, immediately after reset
  // Allow allocation of all timers
  ESP32PWM::allocateTimer(0);
  ESP32PWM::allocateTimer(1);
  ESP32PWM::allocateTimer(2);
  ESP32PWM::allocateTimer(3);

  pinMode(sw0Pin,INPUT_PULLUP); // button switch sw0 Pin
  pinMode(sw1Pin,INPUT_PULLUP); // button switch sw1 Pin
  pinMode(BatPin,INPUT);        // ADC0  battery divider

  setDefaults();        // reset variables

  Wire.begin();         // initialise the I2C interface

  Serial.begin(115200); // baud rate for serial monitor USB Rx/Tx comms
    
  // initialise WiFi link
  WiFi.mode(WIFI_STA);
  // new code for ESP32 lib v3.0+
  while (!WiFi.STA.started()) {delay(10);}
  // Once started we can get the micros MAC address
  myMAC = WiFi.macAddress();
  Serial.println("my MAC= " + String(myMAC));
  // WiFi.disconnect(); // I don't think we need this

  // initialise the ESP-NOW link for the first time
  Init_ESP_NOW();
  
  RunPOST();  // do some basic checks
  
  // set loop timers
  synchLoopTimers();
  
  // define the Core 0 task parameters
  // Core 0 is normally used for WiFi, but here we use it to read sensors
  // leaving Core 1 to create the LED patterns and clock them out
  // Note, this code starts and runs immediately, before loop() does
  xTaskCreatePinnedToCore(
    loop0,       // Task function.
    "Core0",     // name of task.
    10000,       // Stack size of task
    NULL,        // parameter of the task
    1,           // priority of the task
    &Core0,      // Task handle to keep track of created task
    0);          // pin task to core 0
}

// ---------------------------------------------------------------------------------

  // the controller can react to a soft RESET event
  // this function effectively restarts the code as if from power-up
void(* resetFunc) (void) = 0; //declare reset function at address 0

// ---------------------------------------------------------------------------------

void loop() {
  // Main loop code, runs repeatedly.
  //###############################################################################
  //
  //  WiFi comms check & respond
  //
  //###############################################################################
  // if WiFi is conencted then check Rx buffer on every cycle, so that we respond to
  // received packets as soon as possible
  if (RxRecvEvent) {OnDataRecvHandler();}   // respond to ESP-NOW data received events

  if (WiFiRx) {readWiFiRx();}               // buffer contains unread data
    
  readSerial();                             // check serial port on every loop cycle
  
  //###############################################################################
  //
  //  Pan movement tasks
  //
  //###############################################################################
  // perform these tasks at nominally 20ms intervals
  if ((micros() - PanMicros) >= PanInterval) {
    // move the Pan servo towards the target if counter is set
    PanMicros = micros();
    PanMoveEngine(); // perform Pan movement tasks
  }

 
  //###############################################################################
  //
  //  Tilt movement tasks
  //
  //###############################################################################
  // perform these tasks at nominally 20ms intervals
  if ((micros() - TiltMicros) >= TiltInterval) {
    // move the Tilt servo towards the target if counter is set
    TiltMicros = micros();
    TiltMoveEngine(); // perform Tilt movement tasks
  }

  
  //###############################################################################
  //
  //  10ms Tasks
  //
  //###############################################################################
  // perform these tasks at 10ms intervals
  if (Task10ms > 0) {
    switch(Task10ms) {
      case 1: readSW1(); break;  // read right button switch SW1
      case 2: readSW0(); break;  // read left button switch SW0
    } Task10ms--; // count down the sub-tasks from 3 - 1 as we loop
  } else if ((millis() - next10ms) >= 10) {
    // do these every 10ms
    next10ms = millis();  // set next time point
    Task10ms = 2;         // initiate 10ms task sequence
  }
  
  //###############################################################################
  //
  //  20ms Tasks
  //
  //###############################################################################
  // perform these tasks at 20ms intervals
  if (Task20ms > 0) {
    switch(Task20ms) {
      case 1: battRead(); break;            // read analogue input
    } Task20ms--; // count down the sub-tasks from 5 - 1 as we loop
  } else if ((millis() - next20ms) >= 20) {
    // do this every 20ms
    next20ms = millis();  // set next time point
    Task20ms = 1;         // initiate 20ms task sequence

    //###############################################################################
    //
    //  Main tasks - 20ms
    //
    //###############################################################################
    // perform 'Main' tasks in manual button mode
    switch (MainTask) {
      case  0: break; // default null task
      case  1: MoveToVertical(); break;
      case  2: MoveToCalibrate(); break;
      case 99: MoveToRest(); break;
    }
  }

  //###############################################################################
  //
  //  40 ms Tasks (25Hz)
  //
  //###############################################################################
  // These tasks are run in a consequitive ripple to reduce the overhead on more
  // critical code. They are performed in reverse case order and will all complete
  // within the 40ms period
  if (Task40ms > 0) {
    switch(Task40ms) {
      case  1: Display_40ms(); break;          // update display if necessary
      case  2: if (Ping > 0) {Ping--;} break;  // time-out data received flag
      case  3: // if not connected to WiFi run then try to connect every second
        if (!WiFiConnected) {
          // only try to connect WiFi if no 'Pings' are being received
          if (Ping == 0) {
            WiFiConCnt--; if (WiFiConCnt < 1) {
              WiFiConCnt = 10; WiFiTryToConnect();  // 400 ms retry time-out
              if (WiFiTryOnce) {Serial.println("Trying to connect: " + getBAhex());}
            }
          }
        } break;
      case  4:
        if (WiFiPing > 0) {
          WiFiPing--;
          if (WiFiPing < 1) {
             // we have stopped receiving so disconnect WiFi
            WiFiDisconnect();
            ESP_NOW_BA = -1;
            WiFiTryOnce = true;
            Serial.println("WiFiPing == 0");
          }
        } break;
    } Task40ms--;
  } else if ((millis() - next40ms) >= 40) {
    next40ms = millis(); Task40ms = 4;
    // check display mirror auto time-out
    // when count reaches zero DispMon is set to false
    if (DispTx > 0) {DispTx--; if (DispTx == 0) {DispMon = false;}}
  }


  //###############################################################################
  //
  //  40 ms Print Task - Variable
  //
  //###############################################################################
  // this code uses either the serial port or the WiFi link to send messages
  if ((millis() - print40ms) >= 40) {
    print40ms = millis();
    if (!Echo) {PrintTx = "";}    // Echo is false so empty print buffer
    if (PrintTx.length() > 0) {
      // characters in string buffer so send some/all of them
      print40ms = millis() - 34;  // shorten the delay for more data to send
      Tx_PingCnt = 0;             // reset the WiFi ping counter
      if (PrintTgt == 0) {
        // default uses the serial port
        if (Serial.availableForWrite() >= 64) {
          if (PrintTx.length() <= 64) {
            Serial.print(PrintTx);
            PrintTx = "";   // empty the buffer
          } else {
            Serial.print(PrintTx.substring(0,64));
            PrintTx = PrintTx.substring(64);
          }
        }
      } else {
        // text data has been received from WiFi so respond using ESP_NOW
        // ensure that we have had a call back from the previous frame
        if (WiFiConnected) {
          // only trannsmit PrintTx data if connected
          if (WiFiTx_CB == 1) {
            WiFiTx_CB = 0;  // clear the call back flag when sending
            if (PrintTx.length() <= 32) {
              WiFiTx_len = PrintTx.length();
              PrintTx.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
            //  t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = "";   // empty the buffer
            } else {
              WiFiTx_len = 32;
              Any$ = PrintTx.substring(0,WiFiTx_len);
              Any$.toCharArray(Tx_Buff.ESPdata,WiFiTx_len + 1);
              WiFiTxBytes += WiFiTx_len;  // track number of bytes sent
              WiFiTxGetChecksum(WiFiTx_len); Tx_Buff.ESPdata[WiFiTx_len] = WiFiTx_Chk; WiFiTx_len++;
            //  t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
              PrintTx = PrintTx.substring(32);
            }
          } else if (WiFiTx_CB == -1) {
            // an error has occured so resend the data
            if (WiFiTryCnt < 5) {
              WiFiTryCnt++;   // track the number of retries
              WiFiTx_CB = 0;  // clear the call back flag when sending
            //  t0 = micros();
              esp_now_send(broadcastAddress, (uint8_t *) &Tx_Buff, WiFiTx_len);
            } else {
              // too many resends so disconnect WiFi link
              WiFiDisconnect();
              Serial.println("Failed to Tx!");
            }
          }
        } else {
          // not connected so clear PrintTxdata
          PrintTx = "";
        }
      }
    }
  }
}

// ---------------------------------------------------------------------------------

void loop0(void * pvParameters) {
  // this function runs on Core 0, also used for ESP32 WiFi NOW
  // *** WARNING *** do not use for time consuming code like display, I2C, FastLED, etc
  // the following message, sent once at boot, confirms core 0 assignment
  Serial.print("Core0 task running on core "); Serial.println(xPortGetCoreID());

  for(;;){
    // infinite loop, just like loop() but running in Core 0
    // you must never use 'break' or 'return' statement here or it will crash
    // we tell the free RTOS system to call this function every 1 ms
  
  //  t1[2] = micros(); // end time
  //  t0[3] = micros(); // start time

  //###############################################################################
  //
  //  Read light sensors
  //
  //###############################################################################
    // we take a reading and then switch to the next channel, which is read
    // next time this task runs. Each of the 4 sensors are read in sequence
    // if mode Cal == true then use raw values, if false use adjusted
    switch (SensePnt) {
      case 0:
        Sense0Old = Sense0;
        Sense0Raw = 4095 - analogRead(SensePin0);  // take this reading
        Any = analogRead(SensePin1);            // prime this reading
        if (!Cal) {Sense0 = getAdjSense(0,Sense0Raw);}
        else {Sense0 = Sense0Raw;}
        Sense0 = (Sense0 + Sense0Old)/2;
        Sense0Sum -= Sense0Avg; Sense0Sum += Sense0;
        Sense0Avg = Sense0Sum/AvgDiv;
        break;
      case 1:
        Sense1Old = Sense1;
        Sense1Raw = 4095 - analogRead(SensePin1);  // take this reading
        Any = analogRead(SensePin2);            // prime this reading
        if (!Cal) {Sense1 = getAdjSense(1,Sense1Raw);}
        else {Sense1 = Sense1Raw;}
        Sense1 = (Sense1 + Sense1Old)/2;
        Sense1Sum -= Sense1Avg; Sense1Sum += Sense1;
        Sense1Avg = Sense1Sum/AvgDiv;
        break;
      case 2:
        Sense2Old = Sense2;
        Sense2Raw = 4095 - analogRead(SensePin2);  // take this reading
        Any = analogRead(SensePin3);            // prime this reading
        if (!Cal) {Sense2 = getAdjSense(2,Sense2Raw);}
        else {Sense2 = Sense2Raw;}
        Sense2 = (Sense2 + Sense2Old)/2;
        Sense2Sum -= Sense2Avg; Sense2Sum += Sense2;
        Sense2Avg = Sense2Sum/AvgDiv;
        break;
      case 3:
        Sense3Old = Sense3;
        Sense3Raw = 4095 - analogRead(SensePin3);  // take this reading
        Any = analogRead(SensePin0);            // prime this reading
        if (!Cal) {Sense3 = getAdjSense(3,Sense3Raw);}
        else {Sense3 = Sense3Raw;}
        Sense3 = (Sense3 + Sense3Old)/2;
        Sense3Sum -= Sense3Avg; Sense3Sum += Sense3;
        Sense3Avg = Sense3Sum/AvgDiv;
        break;
    }
    // update the pointer, 0 - 3
    SensePnt++; if (SensePnt > 3) {SensePnt = 0;}

  //  t1[3] = micros(); // end time
  //  t0[2] = micros(); // start time

    // return to RTOS system with a 2ms call back, should allow full sensing 125Hz
    vTaskDelay(2/portTICK_PERIOD_MS);
  }
}

// --------------------------------------------------------------------------------

void OnDataRecv(const esp_now_recv_info_t *info, const uint8_t *incomingData, int len) {
  // Call-back function oocurs when data has been received
  // This needs to be treated like an interrupt, so speed is the essence here
  // memcpy moves data in a block, instead of individual bytes if using for() loop
  memcpy(WiFiRxMacAddr, info->src_addr, 6);   // get the senders MAC/broadcast address
  memcpy(WiFiTxMacAddr, info->des_addr, 6);   // get destination MAC address, which should be assigned peer
  memcpy(&Rx_Buff, incomingData, len);        // copy the received data into Rx_Buff array
  Rx_len = len - 1; Rx_Pnt = 0;               // len contain the length of data
  RxRecvEvent = true;                         // set the flag to say data received
}

// --------------------------------------------------------------------------------

void OnDataRecvHandler() {
  // Called from the main loop() when has been set, to handle the received data
  RxRecvEvent = false;
  // incoming data will always be two or more characters, ie. text == $+++
  if (Rx_len < 2) {Serial.println("Rx Error - no Rx data"); return;}

  RxGetChecksum(Rx_len);  // generate a checksum from Rx data
  if (Rx_Chk != Rx_Buff.ESPdata[Rx_len]) {
    Serial.println("Rx Error - checksum");
  }
  // set flags before exiting
  WiFiRxBytes += Rx_len; Rx_Task = 0; WiFiRx = true;
  WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
  WiFiRxRec = true; // set flag indicating data has been recieved from Wii Transceiver
  //  Serial.print("Bytes received: "); Serial.println(Rx_len);
}

// --------------------------------------------------------------------------------

void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status) {
  // Callback when data is sent, determines success or failure
  // t1 = micros();
  // Serial.print("Send Status: ");
  if (WiFiConnected) {
    // the link is already active so handle normal success/failure scenarios
    if (status == 0){
    //  Serial.println("Success " + String(t1-t0));
      WiFiTx = true;    // flag WiFi Tx success
      WiFiTx_CB = 1;    // set flag to indicate that a call back state has occured
      WiFiTxErrCnt = 0; // clear the error flag in case of a glitch
      WiFiPing = 50;    // set link time-out to 2 seconds
    } else {
    //  Serial.println("Fail " + String(t1 - t0));
      WiFiTx = false;   // flag WiFi Tx failure
      WiFiTx_CB = -1;   // set flag to indicate that a resend is required
      WiFiTxErrCnt++;
    }
  } else {
    // not connected so we must have tried to connect and may get an error
    if (status == 0){
      // successful so enter the connected state
      Serial.println("Connected! to " + getBAhex());
      WiFiConnected = true;
      WiFiPing = 50;    // set link time-out to 2 seconds, Wii Transceiver will keep setting this with Wii data
      WiFiTxErrCnt = 0; // reset error count for this connection
    } else {
    //  Serial.println("Status = " + String(status));
      WiFiTxErrCnt++;
    }
    WiFiTx_CB = 1;
  }
}

// ---------------------------------------------------------------------------------

void RunPOST() {
  // start-up initialisation processes  
  setDefaults();  // reset variables
  Serial.println("\n\nLight Tracker v0.0\n");
  Serial.println("Released: " + Released);
  
  // Initialise display
  Display_Init();
  display.setBrightness(Brightness);
  Display_Intro();
  Serial.println("Display initialised");
  DispDel = 50; // wait 2 seconds before displaying anything else

  // set up battery quick averaging
  BatVol = analogRead(BatPin);
  BatAvg = BatVol; BatSum = BatVol * 50;  // preload the battery sum for averaging
  AnyLong = (BatAvg - BatCritical)*100/(BatMax - BatCritical);
  if (AnyLong > 100) {AnyLong = 100;} // limit max to 100%
  if (AnyLong < 0) {AnyLong = 0;}     // limit min to 0%
  Serial.println("Battery " + String(BatVol) + "\t" + String(float(BatAvg)/BatCal) + "v\t" + String(AnyLong) + "%");
  randomSeed(BatVol); // use the battery voltage as the random seed factor
  // if battery is less than 5.5v then USB mode is assumed, which makes servos inoperable
  if (BatAvg <= 2111) {USB = true; Serial.println("USB Mode");}

  // check for SW0 button being pressed when coming out of REST
  if (digitalRead(sw0Pin) == LOW) {
    // SW0 is being pressed so force TEST mode
    TEST = true;
    Display_TestMode();
    Serial.println("TEST mode");
    PanVal = 1500; attachPanServo();
    TiltVal = 1500; attachTiltServo();
  } else {
    // normal reset so move to rest position
    Serial.println("Going to rest");
    SetPanPWM(PanVal);
    SetTiltPWM(TiltVal);
  }

  Serial.println("POST completed");
}

// ---------------------------------------------------------------------------------

void setDefaults() {
  // load default values
  atPanTgt = true;        // = true if Pan servo is at target value
  atTiltTgt = true;       // = true if Tilt servo is at target value
  AvgDiv = 32;            // sensor averaging denominator value
  BatAvg = 4095;          // preload averaging function with high value
  BatAvg = 0;             // preload averaging function with high value
  BatSum = 72320;         // preload averaging sum with 20x max value
  Border = false;         // if true then display has a border
  Bright = 255;           // FastLED brightness
  Brightness = 128;       // Display brightness
  Cal = false;            // == true if in calibration mode
  C_Dn = 0;               // counter for 'C' button down period
  cmdMode = ' ';          // command mode
  cmdSgn = 1;             // cmdVal sign, 1 or -1
  cmdType = ' ';          // command mode type
  cmdVal = 0;             // value associated with a cmdType
  CZ = 0;                 // received button values
  DispBack = false;       // flag used to draw a back arrow on the display
  DispChng = false;       // flag invokes initialisation code
  DispClr = false;        // set == true to clear the memory contents
  DispCnt = 0;            // counter used in display updating
  DispDel = 0;            // >0 display delay counter, stops updates over I2C
  DispDelGbl = 0;         // global DispDel offset, used to adjust display frame rates
  DispDisp = false;       // set = true to immediately display memory contents
  DispLock = false;       // if == true then don't allow display mode changes
  DispMenu = 0;           // group type, 0 = sensors, 1 = tasks, etc
  DispMode = 0;           // the type of display for a given mode
  DispMon = false;        // == true if connected to OLED Mirror app
  DispNext = 0;           // special temp display mode, default = 0
  DispNOP = false;        // set == true to stop display I2C updates
  DispNow = 0;            // if > 0 display special modes
  DispOveride = 0;        // if DispLock == true then use this overide mode
  DispPnt = 0;            // points at the current display mode, default = 0 batteries
  DispTx = 0;             // counter used with display mirror app
  Echo = true;            // if true send print messages
  ErrDiv = 48;            // integrator error scale factor
  ESC = false;            // set = true to escape all functions
  ESP_NOW_BA = -1;        // broadcast address pointer, -1 to start with
  ESP_NOW_Init = false;   // == true once ESP-NOW has been initiaiised
  JoyX = 128;             // received value, default
  JoyY = 128;             // received value, default
  MainTask = 0;           // main task pointer
  PanAtt = false;         // = true when head servo has been attached
  PanInterval = 20000;    // main loop Pan movement interval in microseconds
  PanMax = PanMaxDef;     // maximum Pan servo value
  PanMin = PanMinDef;     // minimum Pan servo value
  PanTgt = PanCtr;        // target Pan servo angle
  PanVal = PanCtr;        // default Pan servo angle in microseconds
  Ping = 0;               // 'ping' counter
  PWM_Freq = 100;         // servo PWM frequency, default = 100;
  Rx_Chk = 0;             // Rx checksum
  RxCZ = 0;               // buffered received CZ value
  RxJoyX = 0;             // buffered received JoyX value
  RxJoyY = 0;             // buffered received JoyY value
  Rx_len = 0;             // length of WiFi received data block
  RxRec = false ;         // set true when a valid frame of data is received
  RxRecvEvent = false;    // set == true when a OnDataRecv() call back event has occured
  RxState = 0;            // receiver state machine state
  Rx_Task = 0;            // pointer user by Rx state machine
  RTxTimeout = 0;         // counter used to auto-reset if WiFi fails
  RxVal = -1;             // value received from serial Rx
  Sense0Avg = 0;          // light sensor averaged readings
  Sense0Avg = 1;          // light sensor averaged readings
  Sense0Avg = 2;          // light sensor averaged readings
  Sense0Avg = 3;          // light sensor averaged readings
  SensePnt = 0;           // pointer to sensor to be read in Core0 task
  Sense0Sum = 0;          // light sensor cumulative readings
  Sense1Sum = 0;          // light sensor cumulative readings
  Sense2Sum = 0;          // light sensor cumulative readings
  Sense3Sum = 0;          // light sensor cumulative readings
  SenseCal = false;       // if == true then sensor averaging has occured
  SerialRx = false;       // == true if any character has been received
  servoPanEn = false;     // false = OFF, true = ON enabled status
  ServoPnt = 0;           // servo target pointer for manual control
  servoTiltEn = false;    // false = OFF, true = ON enabled status
  SubTask = 0;            // sub task pointer
  sw0Cnt = 0;             // button switch counter
  sw0DwnTime = 0;         // button pressed down time
  sw0LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw0_Nop = false;        // == true to block SW0 read tasks, default = false
  sw0State = HIGH;        // state of read button switch pin
  sw0Timer = 0;           // timer used to detemine button sequences
  sw0WaitHi = false;      // set == true to wait for a HIGH state
  sw1Cnt = 0;             // button switch counter
  sw1LastState = HIGH;    // previous state of button switch, HIGH/LOW
  sw1_Nop = false;        // == true to block SW1 read tasks, default = false
  sw1State = HIGH;        // state of read button switch pin
  sw1Timer = 0;           // timer used to detemine button sequences
  sw1WaitHi = false;      // set == true to wait for a HIGH state
  Task10ms = 0;           // 10ms subtask timer
  Task20ms = 0;           // 20ms subtask timer
  Task40ms = 0;           // 40ms subtask timer
  TEST = false;           // normally false, set = true for assembly test mode
  TiltAtt = false;         // = true when head servo has been attached
  TiltInterval = 20000;   // main loop Tilt movement interval in microseconds
  TiltMax = TiltMaxDef;   // maximum Tilt servo value
  TiltMin = TiltMinDef;   // minimum Tilt servo value
  TiltTgt = TiltMax;      // target Tilt servo angle
  TiltVal = TiltMax;      // default Pan servo angle in microseconds
  Tx_PingCnt = 0;         // 1 second Tx ping timer
  USB = false;            // == true if initial battery is below 5.5v
  WiFiCntC = 0;           // C button counter for WiFi enable
  WiFiConCnt = 0;         // counter used in connection retry times
  WiFiConnected = false;  // == true if a WiFi connection has been established
  WiFiEn = false;         // true when WiFi is enabled with prolonged 'Z' button activity
  WiFiPing = 0;           // Rx counter used to maintain connected state
  WiFiRx = false;         // true when a block of data has been received
  WiFiRxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiRxErr;              // number of receive errors in the WiFi link
  WiFiRxRec = false;      // true when a block of data has been successfully received
  WiFiTryCnt = 0;         // WiFi try to connect count
  WiFiTryNum = 0;         // WiFi try to connect total count
  WiFiTryOnce = true;     // == true if first pass of trying to connect
  WiFiTx = false;         // true when a block of data has been sent successfully
  WiFiTxBytes = 0;        // number of bytes receiv ed over the WiFi link
  WiFiTx_CB = 1;          // == 1 call back is clear to send, == -1 resend
  WiFiTx_Chk = 0;         // Tx checksum
  WiFiTx_len = 0;         // >0 when WiFi Tx buffer contains data to be sent
  WiFiTxErr = 0;          // WiFi data error
  WiFiTxErrCnt = 0;       // WiFi Tx error count, leads to disconnection
  WiiType = 0;            // device type, 0 == Nunchuk, 1 == Classic
  Z_Dn = 0;               // button pressed state; 0 = UP, 1 = Down
}

// --------------------------------------------------------------------------------

void synchLoopTimers() {
  // reset or adjust main loop timers, based on zAdvance
  // if  0 then a 2ms phase offset is introduced to prevent simultaneous triggering
  // if >0 then add zAdvance to the current timer values
  while (millis() < 50) {yield();}

  // setting up fresh loop timers
  next10ms = millis() - 10;     // reset 10ms loop timer with 2ms phase to 10ms timer
  next20ms = millis() - 22;     // reset 20ms loop timer with 2ms phase to 10ms timer
  next40ms = millis() - 25;     // reset 40ms loop timer with 2ms phase to 10ms timer
  PanMicros = micros() - 3000;  // reset head movement timer
  print40ms = millis() - 46;    // reset 40ms printer loop timer with 4ms phase to 10ms timer
  TiltMicros = micros() - 3500; // reset head movement timer
}

// ---------------------------------------------------------------------------------
